/*
 * Munge object (.o) files to replace an absolute path for
 * the special "FILE" symbol with a different path, in order
 * to ensure reproducibility against Intel's NixOS docker env.
 *
 * If you don't know ELF, then this blog post will help
 * explain the basics, and thus understand this tool
 *
 * http://blog.k3170makan.com/2018/10/introduction-to-elf-format-part-vi.html
 */

#include <elf.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>

static int debug = 0;

/*
 * @src: original .o FD
 * @hdr: original ELF file header
 * @idx: the index of the section header to read
 * @shdr: fill with the section header
 *
 * Read the section header with index @idx, filling @shdr
 */
int read_section_header(int src, Elf64_Ehdr *hdr, int idx, Elf64_Shdr *shdr)
{
  off_t off = hdr->e_shoff + (idx * sizeof(*shdr));

  if (idx >= hdr->e_shnum) {
    fprintf(stderr, "Section header %d not present\n", idx);
    return -1;
  }

  if (lseek(src, off, SEEK_SET) != off) {
    fprintf(stderr, "Cannot seek to section headers\n");
    return -1;
  }

  if (read(src, shdr, sizeof(*shdr)) != sizeof(*shdr)) {
    fprintf(stderr, "Cannot read section header %d\n", idx);
    return -1;
  }

  return 0;
}

/*
 * @dst: new .o FD
 * @shdr: the section header for section being copied
 * @idx: the index of the section header to write
 * @shdr: the section header to write
 *
 * Write the section header with index @idx
 */
int write_section_header(int dst, Elf64_Ehdr *hdr, int idx, Elf64_Shdr *shdr)
{
  off_t off = hdr->e_shoff + (idx * sizeof(*shdr));

  if (idx >= hdr->e_shnum) {
    fprintf(stderr, "Section header %d not present\n", idx);
    return -1;
  }

  if (lseek(dst, off, SEEK_SET) != off) {
    fprintf(stderr, "Cannot seek to section headers\n");
    return -1;
  }

  if (write(dst, shdr, sizeof(*shdr)) != sizeof(*shdr)) {
    fprintf(stderr, "Cannot write section header %d: %s\n", idx, strerror(errno));
    return -1;
  }

  return 0;
}

/*
 * @src: original .o FD
 * @dst: new .o FD
 * @shdr: the section header for section being copied
 *
 * Copy the section data for @shdr, moving its location
 * by @reduce. Modify @shdr to refer to the new location
 */
int copy_section(int src, int dst, Elf64_Shdr *shdr, ssize_t reduce)
{
  char data[1024*64];
  size_t done = 0;

  if (shdr->sh_size == 0)
    return 0;

  if (lseek(src, shdr->sh_offset, SEEK_SET) != (off_t)shdr->sh_offset) {
    fprintf(stderr, "Cannot seek to section\n");
    return -1;
  }

  if (lseek(dst, shdr->sh_offset - reduce, SEEK_SET) != (off_t)(shdr->sh_offset - reduce)) {
    fprintf(stderr, "Cannot seek to section\n");
    return -1;
  }

  while (done < shdr->sh_size) {
    ssize_t want = shdr->sh_size - done;
    if (want > (ssize_t)sizeof(data)) {
      want = sizeof(data);
    }

    if (read(src, &data, want) != want) {
      fprintf(stderr, "Cannot read section at %zu (%zu bytes)\n", shdr->sh_offset, shdr->sh_size);
      return -1;
    }

    if (write(dst, &data, want) != want) {
      fprintf(stderr, "Cannot write section at %zu (%zu bytes)\n", shdr->sh_offset, shdr->sh_size);
      return -1;
    }

    done += want;
  }

  shdr->sh_offset -= reduce;
  
  return 0;
}

char *read_str(int src, Elf64_Shdr *strtab, off_t off)
{
  char str[1024];

  lseek(src, strtab->sh_offset + off, SEEK_SET);

  read(src, str, sizeof(str));

  if (memchr(str, '\0', sizeof(str)) == NULL)
    return NULL;

  return strdup(str);
}

/*
 * @src: original .o FD
 * @hdr: original ELF file header
 * @shstrtab: section header for .shstrtab
 * @type: the section type to locate
 * @wantname: the section name to locate
 * @shdr: fill with section header
 *
 * Find the section header for the section called @wantname
 * with type @type, and fill in @shdr
 */
int get_section(int src, Elf64_Ehdr *hdr, Elf64_Shdr *shstrtab, Elf64_Word type, const char *wantname, Elf64_Shdr *shdr)
{
  int i;

  for (i = 0; i < hdr->e_shnum; i++) {
    char *name;
    if (read_section_header(src, hdr, i, shdr) < 0)
      return -1;

    if (!(name = read_str(src, shstrtab, shdr->sh_name))) {
      fprintf(stderr, "Missing section name %d\n", i);
      return -1;
    }

    if (strcmp(name, wantname) == 0 &&
        type == shdr->sh_type) {
      free(name);
      return 0;
    }

    free(name);
  }

  fprintf(stderr, "Section '%s' with type %d, not found\n", wantname, type);
  return -1;
}


/*
 * @src: original .o FD
 * @symtab: section header for .symtab
 * @sym: fill with the FILE symbol data
 *
 * Locate the "FILE" symbol referenced by @symtab, and
 * fill in @sym with its structure
 */
int get_file_sym(int src, Elf64_Shdr *symtab, Elf64_Sym *sym)
{
  int i;
  int nsym = (symtab->sh_size / sizeof(Elf64_Sym));

  if ((symtab->sh_size % sizeof(Elf64_Sym)) != 0) {
    fprintf(stderr, ".symtab size is not a multiple of symbol entry size\n");
    return -1;
  }

  for (i = 0; i < nsym; i++) {
    off_t off = symtab->sh_offset + (sizeof(Elf64_Sym) * i);

    if (lseek(src, off, SEEK_SET) != off) {
      fprintf(stderr, "Cannot seek to symbol %d at %zu\n", i, off);
      return -1;
    }

    if (read(src, sym, sizeof(*sym)) != sizeof(*sym)) {
      fprintf(stderr, "Cannot read symbol %d at %zu\n", i, off);
      return -1;
    }

    if (sym->st_info == STT_FILE) {
      return 0;
    }
  }

  fprintf(stderr, "File symbol not found\n");
  return -1;
}

/*
 * @src: original .o FD
 * @dst: new .o FD
 * @shdr: section header for .strtab
 * @offset: offset of string that is going to be changed
 * @delta: difference in length of @srcdir vs @dstdir
 *
 * Munge the .symtab, changing location of any strings after
 * @offset, to be moved by @delta.
 */
int munge_symtab(int src, int dst, Elf64_Shdr *strtab, Elf64_Shdr *shdr, off_t offset, ssize_t delta)
{
  int i;
  int nsym = (shdr->sh_size / sizeof(Elf64_Sym));

  for (i = 0; i < nsym; i++) {
    Elf64_Sym sym;
    off_t off = shdr->sh_offset + (sizeof(Elf64_Sym) * i);

    if (lseek(src, off, SEEK_SET) != off) {
      fprintf(stderr, "Cannot seek to symbol %d at %zu\n", i, off);
      return -1;
    }
    if (lseek(dst, off, SEEK_SET) != off) {
      fprintf(stderr, "Cannot seek to symbol %d at %zu\n", i, off);
      return -1;
    }

    if (read(src, &sym, sizeof(sym)) != sizeof(sym)) {
      fprintf(stderr, "Cannot read symbol %d at %zu\n", i, off);
      return -1;
    }

    char *name = read_str(src, strtab, sym.st_name);

    if (sym.st_name > offset) {
      if (debug)
        fprintf(stderr, "Munge sym %d '%s' %d -> %zu\n", i, name, sym.st_name, sym.st_name - delta);
      sym.st_name -= delta;

      if (sym.st_info == STT_FILE)
        sym.st_size -= delta;
    } else {
      if (debug)
        fprintf(stderr, "Copy sym %d '%s' %d\n", i, name, sym.st_name);
    }

    if (write(dst, &sym, sizeof(sym)) != sizeof(sym)) {
      fprintf(stderr, "Cannot write symbol %d at %zu\n", i, off);
      return -1;
    }
  }

  return 0;
}

/*
 * @src: original .o FD
 * @dst: new .o FD
 * @shdr: section header for .strtab
 * @offset: offset of string to munge
 * @delta: difference in length of @srcdir vs @dstdir
 * @srcdir: original FILE path to remove
 * @dstdir: new FILE path to add
 *
 * Munge the .strtab string the starts at @offset, to remove @srcdir
 * and add @dstdir
 */
int munge_strtab(int src, int dst, Elf64_Shdr *shdr, off_t offset, ssize_t delta, const char *srcdir, const char *dstdir)
{
  char *oldstrs;
  char *newstrs;

  if (lseek(src, shdr->sh_offset, SEEK_SET) != (off_t)shdr->sh_offset) {
    fprintf(stderr, "Cannot seek to .strtab at %zu\n", shdr->sh_offset);
    return -1;
  }

  if (lseek(dst, shdr->sh_offset, SEEK_SET) != (off_t)shdr->sh_offset) {
    fprintf(stderr, "Cannot seek to .strtab at %zu\n", shdr->sh_offset);
    return -1;
  }

  oldstrs = malloc(shdr->sh_size);
  if (read(src, oldstrs, shdr->sh_size) != (ssize_t)shdr->sh_size) {
    fprintf(stderr, "Cannot read .strtab at %zu size %zu\n", shdr->sh_offset, shdr->sh_size);
    free(oldstrs);
    return -1;
  }

  newstrs = malloc(shdr->sh_size - delta);
  memmove(newstrs, oldstrs, offset);
  memmove(newstrs + offset, dstdir, strlen(dstdir));
  memmove(newstrs + offset + strlen(dstdir),
          oldstrs + offset + strlen(srcdir),
          shdr->sh_size - (offset + strlen(srcdir)));

  if (debug)
    fprintf(stderr, "Munge strtab size %zu -> %zu\n", shdr->sh_size, shdr->sh_size - delta);
  shdr->sh_size -= delta;

  if (write(dst, newstrs, shdr->sh_size) != (ssize_t)shdr->sh_size) {
    fprintf(stderr, "Cannot write .strtab at %zu size %zu\n", shdr->sh_offset, shdr->sh_size);
    return -1;
  }

  return 0;
}

/*
 * @src: original .o FD
 * @dst: new .o FD
 * @hdr: original ELF file header
 * @srcdir: original FILE path to remove
 * @dstdir: new FILE path to add
 *
 * Copy all ELF sections, munging the .symtab and .strtab sections to
 * change the FILE symbol path
 */
int munge_sections(int src, int dst, Elf64_Ehdr *hdr, const char *srcdir, const char *dstdir)
{
  int i;
  Elf64_Shdr shstrtab;
  Elf64_Shdr symtab;
  Elf64_Shdr strtab;
  Elf64_Sym filesym;
  ssize_t delta = strlen(srcdir) - strlen(dstdir);
  ssize_t reduce = 0;

  if (read_section_header(src, hdr, hdr->e_shstrndx, &shstrtab) < 0)
    return -1;

  if (get_section(src, hdr, &shstrtab, SHT_SYMTAB, ".symtab", &symtab) < 0)
    return -1;

  if (get_section(src, hdr, &shstrtab, SHT_STRTAB, ".strtab", &strtab) < 0)
    return -1;

  if (get_file_sym(src, &symtab, &filesym) < 0)
    return -1;

  char *name = read_str(src, &strtab, filesym.st_name);

  if (strncmp(srcdir, name, strlen(srcdir)) != 0) {
    fprintf(stderr, "File '%s' does not start with prefix '%s'\n", name, srcdir);
    free(name);
    return -1;
  }
  free(name);

  for (i = 0; i < hdr->e_shnum; i++) {
    Elf64_Shdr shdr;

    if (read_section_header(src, hdr, i, &shdr) < 0)
      return -1;

    char *name = read_str(src, &shstrtab, shdr.sh_name);

    if (strcmp(name, ".symtab") == 0) {
      free(name);

      if (debug)
        fprintf(stderr, "Munge .symtab (%d)\n", filesym.st_name);
      if (munge_symtab(src, dst, &strtab, &shdr, filesym.st_name, delta) < 0)
        return -1;

    } else if (strcmp(name, ".strtab") == 0) {
      ssize_t pad = 0;
      pad = shdr.sh_size % 16 ? 16 - (shdr.sh_size % 16) : 0;
      reduce = shdr.sh_size + pad;
      free(name);

      if (debug)
        fprintf(stderr, "Munge .strtab\n");
      if (munge_strtab(src, dst, &shdr, filesym.st_name, delta, srcdir, dstdir) < 0)
        return -1;

      pad = shdr.sh_size % 16 ? 16 - (shdr.sh_size % 16) : 0;
      reduce -= (shdr.sh_size + pad);
    } else {
      if (debug)
        fprintf(stderr, "Copy %s (reduce %zu)\n", name, reduce);

      free(name);

      if (copy_section(src, dst, &shdr, reduce) < 0)
        return -1;
    }
    
    if (write_section_header(dst, hdr, i, &shdr) < 0)
      return -1;
  }
  
  return 0;
}

/*
 * @src: original .o FD
 * @dst: new .o FD
 * @srcdir: original FILE path to remove
 * @dstdir: new FILE path to add
 *
 * Open the ELF file, read the header, validate it,
 * write the new header, and then copy sections
 */
int munge(int src, int dst, const char *srcdir, const char *dstdir)
{
  Elf64_Ehdr hdr;
  off_t end;

  if (read(src, &hdr, sizeof(hdr)) != sizeof(hdr)) {
    fprintf(stderr, "Cannot read header\n");
    return -1;
  }

  if (memcmp(hdr.e_ident, ELFMAG, 4) != 0) {
    fprintf(stderr, "Wrong magic bytes\n");
    return -1;
  }

  if (hdr.e_type != ET_REL) {
    fprintf(stderr, "Expected object file\n");
    return -1;
  }
   
  if (write(dst, &hdr, sizeof(hdr)) != sizeof(hdr)) {
    fprintf(stderr, "Cannot write header\n");
    return -1;
  }

  if (hdr.e_phoff != 0 || hdr.e_phnum != 0) {
    fprintf(stderr, "Unexpected program headers present\n");
    return -1;
  }

  if (hdr.e_shentsize != sizeof(Elf64_Shdr)) {
    fprintf(stderr, "Unexpected section header size %d != %zu\n",
            hdr.e_shentsize, sizeof(Elf64_Shdr));
    return -1;
  }

  if (munge_sections(src, dst, &hdr, srcdir, dstdir) < 0)
    return -1;

  end = lseek(dst, 0, SEEK_END);
  end += end % 16 ? 16 - (end % 16) : 0;
  if (ftruncate(dst, end) < 0) {
    fprintf(stderr, "Cannot pad to %zu\n", end);
    return -1;
  }

  return 0;
}

int main(int argc, char **argv)
{
  char *srcdir;
  char *dstdir;
  char *obj;
  char *newobj;
  int src = -1;
  int dst = -1;
  int rv = -1;
  if (argc != 4) {
    fprintf(stderr, "syntax: %s SRC-PREFIX DST-PREFIX OBJ-FILE\n", argv[0]);
    exit(EXIT_FAILURE);
  }

  if (getenv("MUNGEELF_DEBUG"))
    debug = 1;

  srcdir = argv[1];
  dstdir = argv[2];
  obj = argv[3];

  asprintf(&newobj, "%s.mod", obj);

  unlink(newobj);

  if ((src = open(obj, O_RDONLY)) < 0) {
    fprintf(stderr, "Cannot read %s\n", obj);
    goto cleanup;
  }
  if ((dst = open(newobj, O_WRONLY | O_CREAT | O_TRUNC, 0666)) < 0) {
    fprintf(stderr, "Cannot create %s\n", obj);
    goto cleanup;
  }

  if (munge(src, dst, srcdir, dstdir) < 0)
    goto cleanup;

  if (rename(newobj, obj) < 0) {
    fprintf(stderr, "Cannot rename %s -> %s\n", newobj, obj);
    goto cleanup;
  }

  fprintf(stderr, "Re-wrote %s\n", obj);
  rv = 0;
  
 cleanup:
  if (src != -1)
    close(src);
  if (dst != -1)
    close(dst);
  
  if (rv == -1)
    unlink(newobj);

  exit(rv == 0 ? EXIT_SUCCESS : EXIT_FAILURE);
}
